import pathlib
from string import Template
from collections import namedtuple
from typing import Dict
from .constants import ASTEROID_TEMPLATE_PATH, ADDITIONAL_TEMPLATE_PATH
from .functions import _convert_input_to_fortran_strings
from .mesh import PySALEMesh


TimeStep = namedtuple('TimeStep', ['initial', 'max', 'end', 'save'])


class InputFile:
    """Base input file class.

    """
    def __init__(self, type_: str):
        """Input file type must be supplied to the constructor.

        Parameters
        ----------
        type_ : str
        """
        self.type = type_
        self._template = None
        self._template_path = None

    @property
    def template_path(self) -> pathlib.Path:
        """Path to the template

        Returns
        -------
        path : pathlib.Path
        """
        if self._template_path is None:
            if self.type == 'asteroid':
                self._template_path = ASTEROID_TEMPLATE_PATH
            elif self.type == 'additional':
                self._template_path = ADDITIONAL_TEMPLATE_PATH
            else:
                raise TypeError(f'Invalid input file type supplied. '
                                f'Must be "asteroid" or "additional", '
                                f'not {self.type}')
        return self._template_path

    @property
    def template(self) -> Template:
        """The template string for the input file.

        Returns
        -------
        template : str
        """
        if self._template is None:
            with open(pathlib.Path(self.template_path), 'r') as f:
                self._template = Template(f.read())
        return self._template


class AsteroidInput(InputFile):
    """asteroid.inp file creator class.

    Allows a user to build an asteroid.inp file and uses a simple
    template to write one to a new file of the user's choice.
    """
    def __init__(self, model_name: str, timestep: TimeStep, mesh: PySALEMesh):
        """Construct the AsteroidInput class.

        Only model_name, timestep and mesh, need to be provided.
        However, multiple other properties are also set. These are:

            * surface_temperature = 300.
            * east_bc = 'FREESLIP'
            * west_bc = 'FREESLIP'
            * south_bc = 'OUTFLOW'
            * north_bc = 'OUTFLOW'

        But they can all be changed after initialisation.

        Parameters
        ----------
        model_name : str
        timestep : TimeStep
        mesh : PySALEMesh
        """
        super().__init__('asteroid')
        self.model_name = model_name
        self.timestep = timestep
        self.mesh = mesh
        self._template = None
        self.surface_temperature = 300.
        self.east_bc = 'FREESLIP'
        self.west_bc = 'FREESLIP'
        self.south_bc = 'OUTFLOW'
        self.north_bc = 'OUTFLOW'

    def _parse_mesh(self):
        zones = {'east': 0, 'west': 0, 'north': 0, 'south': 0}
        for z in self.mesh.extension_zones:
            zones[z.region.name.lower()] = z.depth
        return zones['east'], zones['west'], zones['north'], zones['south']

    def _perform_substitutions(self):
        east, west, north, south = self._parse_mesh()
        output = self.template.substitute(
            _convert_input_to_fortran_strings(
                {
                    'simulation_name': self.model_name,
                    'east_cells': east,
                    'west_cells': west,
                    'north_cells': north,
                    'south_cells': south,
                    'cell_size': self.mesh.cell_size,
                    'high_res_x': self.mesh.x,
                    'high_res_y': self.mesh.y,
                    'cyl_geometry': int(self.mesh.cylindrical_symmetry),
                    'collision_index': self.mesh.collision_site,
                    'objresh': self.mesh.objresh,
                    'vert_offset': self.mesh.vertical_offset,
                    'timestep': self.timestep.initial,
                    'timestep_max': self.timestep.max,
                    'time_end': self.timestep.end,
                    'save_interval': self.timestep.save,
                    'east_boundary_condition': self.east_bc,
                    'west_boundary_condition': self.west_bc,
                    'north_boundary_condition': self.north_bc,
                    'south_boundary_condition': self.south_bc,
                    'max_cell_size': self.mesh.extension_factor.max_cell_size,
                    'ext_factor': self.mesh.extension_factor.multiplier,
                    'surface_temperature': self.surface_temperature
                }
            )
        )
        return output

    def write_to(self,
                 path: pathlib.Path =
                 pathlib.Path.cwd() / 'asteroid.inp') -> None:
        """Write the asteroid.inp file to given path.

        Defaults to asteroid.inp in the current working directory.

        Parameters
        ----------
        path : pathlib.Path
        """
        with open(path, 'w') as f:
            f.write(self._perform_substitutions())


class AdditionalInput(InputFile):
    """additional.inp file creator class.

    Allows a user to build an additional.inp file and uses a simple
    template to write one to a new file of the user's choice.
    """
    def __init__(self, mesh: PySALEMesh,
                 material_names: Dict[int, str],
                 host_object_number: int = 1):
        """Initialise the AdditionalInput class.

        Parameters
        ----------
        mesh : PySALEMesh
        material_names : Dict[int, str]
        host_object_number : int
        """
        super(AdditionalInput, self).__init__('additional')
        self.mesh = mesh
        self.material_names = material_names
        self.host_number = host_object_number

    def _perform_substitutions(self):
        material_names = ' : '.join(self.material_names[i]
                                    for i in self.mesh.material_numbers)
        host_numbers = ' : '.join(str(self.host_number)
                                  for i in self.mesh.material_numbers)
        output = self.template.substitute(
            _convert_input_to_fortran_strings(
                {
                    'material_number': len(self.mesh.material_numbers),
                    'material_names': material_names,
                    'host_object_numbers': host_numbers
                }
            )
        )
        return output

    def write_to(self,
                 path: pathlib.Path =
                 pathlib.Path.cwd() / 'asteroid.inp') -> None:
        """Write the asteroid.inp file to given path.

        Defaults to additional.inp in the current working directory.

        Parameters
        ----------
        path : pathlib.Path
        """
        with open(path, 'w') as f:
            f.write(self._perform_substitutions())



import pathlib
import random
from .objects import PySALEObject
from .constants import GRAIN_LIBRARY_PATH


class GrainLibrary:
    """Loads and deploys grains generated from file."""
    def __init__(self):
        self.grains = []

    def load_from_directory(self, directory: pathlib.Path) -> None:
        """Given a directory, loads all grain files it can find.

        Parameters
        ----------
        directory : pathlib.Path

        """
        for file in directory.iterdir():
            if file.is_file():
                try:
                    grain = PySALEObject.create_from_file(file)
                    self.grains.append(grain)
                except Exception as e:
                    print(e)

    def load_builtin_library(self) -> None:
        """Load all grains in the builtin library"""
        self.load_from_directory(GRAIN_LIBRARY_PATH)

    def random_grain(self):
        """Return a random grain from the list.

        Returns
        -------
        grain : PySALEObject
        """
        return random.choice(self.grains)



from abc import ABC
from math import pi, sqrt
from shapely.geometry import Polygon, Point
import shapely.affinity as affinity
import matplotlib.pyplot as plt
from typing import Union, Dict, List, Optional, Tuple
from pathlib import Path
from collections import namedtuple
import numpy as np


Velocity = namedtuple('Velocity', ['x', 'y'])


class PySALEObject(Polygon, ABC):
    """Base object for all objects in PySALESetup

    Based on `shapely.geometry.Polygon`.

    Examples
    --------
    This object forms the basis of everything we do in PySALESetup.
    First here's an example of a rectangular object 10 m x 15 m.

    >>> from PySALESetup import PySALEObject
    >>> import matplotlib.pyplot as plt
    >>> main = PySALEObject([(0, 0), (0, 15), (10, 15), (10, 0)])
    >>> main.plot()
    >>> plt.show()

    This example creates a triangular object.

    >>> from PySALESetup import PySALEObject
    >>> import matplotlib.pyplot as plt
    >>> main = PySALEObject([(0, 0), (5, 5), (10, 0)])
    >>> main.plot()
    >>> plt.show()

    This example creates an elliptical object with centroid (5, 5)
    major axis 1, minor axis 0.5 and rotated by 10 degrees.

    >>> from PySALESetup import PySALEObject
    >>> import matplotlib.pyplot as plt
    >>> main = PySALEObject.generate_ellipse([5., 5.], 1., .5, 10.)
    >>> main.plot()
    >>> plt.show()

    """
    def __init__(self, *args, **kwargs):
        """Construct a Polygon PySALEObject instance.

        Parameters
        ----------
        args
            See shapely.geometry.Polygon args
        kwargs
            See shapely.geometry.Polygon kwargs
        """
        super(PySALEObject, self).__init__(*args, **kwargs)
        self._children = None
        self._material_colors = None
        self._velocity = Velocity(x=0., y=0.)
        if 'material' in kwargs:
            self._material = kwargs['material']
        else:
            self._material = None

    def __repr__(self):
        coords = [(i, j) for i, j in zip(*self.exterior.xy)]
        return f"<PySALEObject({coords})>"

    def copy_properties_to_new_polygon(self, polygon: 'PySALEObject') \
            -> 'PySALEObject':
        """Copy properties to a another polygon.

        Properties copied are:
            - children
            - velocity
            - material colors
            - material number

        Parameters
        ----------
        polygon : PySALEObject

        Returns
        -------
        target : PySALEObject
        """
        polygon._children = self._children[:] \
            if self._children is not None else None
        polygon._velocity = self._velocity
        polygon._material_colors = self._material_colors
        polygon._material = self._material
        return polygon

    @classmethod
    def generate_ellipse(cls, xy: List[float],
                         major: float,
                         minor: float,
                         rotation: float,
                         material: int = 1) -> 'PySALEObject':
        """Create an "ellipse" version of the PySALEObject.

        ellipse PySALEObjects are actually just polygons with many
        points (65) and not true ellipses, however, given that these
        are to be applied to a discrete mesh, exact ellipses are not
        necessary.

        Parameters
        ----------
        xy : List[float]
            centroid of the ellipse
        major : float
            major radius/axis of the ellipse
        minor : float
            minor radius/axis of the ellipse
        rotation : float
            rotation of the ellipse in degrees from the horizontal
        material : int
            material number to be assigned to. Defaults to 1.

        Returns
        -------
        ellipse : PySALEObject
        """
        ellipse = Point(*xy).buffer(1)
        ellipse = affinity.scale(ellipse, float(major), float(minor))
        ellipse = affinity.rotate(ellipse, rotation) \
            if rotation > 0 else ellipse
        ellipse_coords = [(i, j) for i, j in zip(*ellipse.exterior.xy)]
        new = cls(ellipse_coords)
        new.set_material(material)
        return new

    @classmethod
    def create_from_file(cls, file_name: Path) -> 'PySALEObject':
        """Create a polygon object from coordinates in a csv file.

        It is assumed the text file is of the following format.

        > 0., 0.
        > 0., 1.
        > 1., 1.
        > 1., 0.

        With no headers or footers.

        This function is designed to be used with the grain library.

        Parameters
        ----------
        file_name : Path

        Returns
        -------
        new : PySALEObject
        """
        coords = np.genfromtxt(file_name, delimiter=',')
        new = cls(coords)
        return new

    @property
    def children(self) -> 'List[PySALEObject]':
        """PySALEObjects contained (and spawned from) the PySALEObject.

        Returns
        -------
        children : List[PySALEObject]
        """
        if self._children is None:
            self._initialise_children()
        return self._children

    @property
    def velocity(self) -> Velocity:
        """Named tuple Velocity assigned to the object.

        Use `PySALEObject.set_velocity()` to change this.

        Returns
        -------
        velocity_object : Velocity
        """
        return self._velocity

    @property
    def is_void(self) -> bool:
        """Return a bool indicating if the polygon is a void.

        Returns
        -------
        void_status : bool
        """
        return self._material == 0

    @property
    def material(self) -> int:
        """Returns the material assigned to the polygon.

        If no material has yet been assigned, it is assigned to material
        number 1.

        Returns
        -------
        material_number : int
        """
        if self._material is None:
            self.set_material(1)
        return self._material

    @property
    def material_colors(self) -> Dict[int, str]:
        """Returns the colors assigned to each material number.

        Returns
        -------
        material_colors : Dict[int, str]
        """
        if self._material_colors is None:
            self.set_material_colormap('viridis')
        return self._material_colors

    def set_velocity(self, x: float, y: float,
                     include_children: bool = True) -> None:
        """Set the velocity of the object.

        Parameters
        ----------
        x : float
        y : float
        include_children : bool
            Set velocity of all children as well? defaults to True.

        Returns
        -------
        None
        """
        self._velocity = Velocity(x=x, y=y)
        for child in self.children:
            child.set_velocity(x, y)

    def set_as_void(self) -> None:
        """Set the material to void.

        Additionally sets the velocities to 0.0.

        Returns
        -------
        None
        """
        self.set_velocity(0., 0.)
        self._material = 0

    def set_material(self, material: int) -> None:
        """Set the material number of the object.

        material must be between 0 and 9 inclusive.
        0 represents void.

        Parameters
        ----------
        material : int

        Returns
        -------
        None
        """
        assert type(material) == int and 0 <= material <= 9,\
            f"material number must be a number " \
            f"from 0 - 9. Not {material}."
        if material == 0:
            self.set_as_void()
        else:
            self._material = material

    def calculate_equivalent_radius(self):
        """Calculate the radius of the circle with the same area.

        Returns
        -------
        float
            Radius of the equivalent circle
        """

        return sqrt(self.area/pi)

    @property
    def has_children(self) -> bool:
        """Returns True if object has children.

        Returns
        -------
        has_children : bool
        """
        if self._children is not None:
            return True
        else:
            return False

    def _initialise_children(self):
        if self._children is None:
            self._children = []

    def add_child(self, child: 'PySALEObject') -> None:
        """Add existing object as a child to this object.

        Parameters
        ----------
        child : PySALEObject

        Returns
        -------
        None
        """
        self._initialise_children()
        self._children.append(child)

    def spawn_polygon_in_shape(self, *args, **kwargs) -> 'PySALEObject':
        """Create a child polygon for the object.

        Child must be contained within the host, otherwise an
        AssertionError is raised.

        Parameters
        ----------
        args
        kwargs

        Returns
        -------
        polygon : PySALEObject
        """
        child = PySALEObject(*args, **kwargs)
        self._process_new_child(child)
        return child

    def _process_new_child(self, child) -> None:
        assert self.contains(child), \
            "child objects must remain within the host"
        self._initialise_children()
        self._children.append(child)

    def spawn_ellipse_in_shape(self, xy: List[float],
                               major: float,
                               minor: float,
                               rotation: float,
                               material: int = 1) -> 'PySALEObject':
        """Create a child ellipse for the object.

        Parameters
        ----------
        xy : List[float]
            coordinates of the centroid
        major : float
            major axis
        minor : float
            minor axis
        rotation : float
            angle to the horizontal, anticlockwise, in degrees
        material : int
            material for the spawned object to have. Defaults to 1.

        Returns
        -------
        polygon : PySALEObject
        """
        ellipse = PySALEObject.generate_ellipse(xy, major, minor,
                                                rotation, material)
        self._process_new_child(ellipse)
        return ellipse

    def scale_object(self, factor: float,
                     area: bool = False) -> 'PySALEObject':
        """Create new scaled object from current object.

        Parameters
        ----------
        factor : float
        area : bool
            Does the factor refer to area or equivalent radius?

        Returns
        -------
        PySALEObject
            The new scaled object
        """
        if area is True:
            new_area = self.area * abs(factor)
            old_radius = self.calculate_equivalent_radius()
            new_radius = sqrt(new_area/pi)
            factor = new_radius/old_radius
        return self.resize(factor, factor)

    def plot(self,
             ax: Optional[plt.Axes] = None,
             include_children: bool = True,
             cmap: str = 'rainbow') -> Tuple[plt.Figure, plt.Axes]:
        """Plot the object on matplotlib axes.

        If no axes, the method will create some.

        Parameters
        ----------
        ax : Optional[plt.Axes]
        include_children : bool
            Recursively plot all children as well
        cmap : str
            The colormap used when plotting

        Returns
        -------
        fig, ax : plt.Figure, plt.Axes
            If axes were provided these are the same that were given.
        """
        self.set_material_colormap(cmap)
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        else:
            fig = ax.get_figure()
        x, y = self.exterior.xy
        if self.material is not None:
            ax.plot(x, y, color=self.material_colors[self.material])
        else:
            ax.plot(x, y, color=self.material_colors[self.material])
        if include_children and self.has_children:
            for child in self._children:
                child.plot(ax, include_children, cmap)
        ax.set_aspect(1)
        return fig, ax

    def set_material_colormap(self,
                              colormap: str,
                              void: str = 'gray',
                              unassigned: str = 'brown') -> None:
        """Set the colors to be used when plotting.

        Parameters
        ----------
        colormap : str
        void : str
        unassigned : str

        Returns
        -------
        None
        """
        cm = plt.get_cmap(colormap)
        self._material_colors = {i: cm(1.*i/9.) for i in range(1, 9+1)}
        self._material_colors[0] = void
        self._material_colors[-1] = unassigned

    @property
    def krumbein_phi(self) -> float:
        """Return the Krumbein Phi of the object.

        A dimensionless, logarithmic, grain size metric commonly used in
        particle size distributions.

        >>> krumbein_phi = -log2(D/D0)

        Where D is the diameter of the sphere (circle in our case)
        that has the same volume (area) as the grain in question,
        and D0 is a reference length equal to 1 mm (so that the
        equation remains dimensionally consistent.

        See [Investigating the effect of mesostructure on
        the shock response of granular materials
        through numerical modelling
        - James G. Derrick 2018
        Section 6.1](https://doi.org/10.25560/66097)

        Returns
        -------
        krumbein_phi : float
        """
        diameter = np.sqrt(self.area/np.pi)*2.
        krumbein_phi = -np.log2(diameter)
        return krumbein_phi

    @property
    def elongation(self) -> float:
        length_bbox = max((self.bounds[2] - self.bounds[0]),
                          (self.bounds[3] - self.bounds[1]))
        width_bbox = min((self.bounds[2] - self.bounds[0]),
                         (self.bounds[3] - self.bounds[1]))
        elongation = length_bbox/width_bbox
        return elongation

    @property
    def area_ratio(self):
        length_bbox = max((self.bounds[2] - self.bounds[0]),
                          (self.bounds[3] - self.bounds[1]))
        width_bbox = min((self.bounds[2] - self.bounds[0]),
                         (self.bounds[3] - self.bounds[1]))
        area_ratio = (length_bbox*width_bbox)/self.area
        return area_ratio

    def translate(self, newx: float, newy: float):
        """Translate from one centroid to another.

        Returns a copy of the translated polygon.
        Does NOT change the polygon itself.

        Parameters
        ----------
        newx : float
        newy : float

        Returns
        -------
        translated_copy : PySALEObject
        """
        centroid = [self.centroid.x, self.centroid.y]
        diffs = [a - centroid[i] for i, a in enumerate([newx, newy])]
        new = self.copy_properties_to_new_polygon(
            affinity.translate(self, *diffs)
        )

        new._children = [child.translate(newx, newy)
                         for child in self.children]
        return new

    def rotate(self, angle: float,
               origin: Union[str, Tuple[float, float], Point] = 'center'):
        """Rotate by `angle` degrees and about the point `origin`.

        Returns a copy of the rotated polygon.
        Does NOT change the polygon itself.


        Parameters
        ----------
        angle : float
            rotation amount in degrees anticlockwise from the horizontal
        origin : Union[str, Point, Tuple[float, float]]
            can either be the string 'center', where the self origin
            is used or it can be a shapely.geometry.Point object.

        Returns
        -------
        rotated_copy : PySALEObject
        """
        new = self.copy_properties_to_new_polygon(
            affinity.rotate(self, angle, origin=origin)
        )
        if origin == 'center':
            new._children = [child.rotate(angle, origin=self.centroid)
                             for child in self.children]
        else:
            if isinstance(origin, tuple):
                origin = Point(*origin)
            new._children = [child.rotate(angle, origin=origin)
                             for child in self.children]
        return new

    def resize(self, xfactor: float = 1., yfactor: float = 1.):
        """Resize by xfactor and yfactor in the x and y directions.

        Returns a copy of the resized polygon.
        Does NOT change the polygon itself.

        Parameters
        ----------
        xfactor : float
        yfactor : float

        Returns
        -------
        resize_copy : PySALEObject
        """
        new = self.copy_properties_to_new_polygon(
            affinity.scale(self, xfactor, yfactor)
        )
        new._children = [child.resize(xfactor, yfactor)
                         for child in self.children]
        return new



from PySALESetup.objects import PySALEObject
import warnings
from typing import Tuple, List, Callable
from shapely.geometry import MultiPolygon
import numpy as np
import scipy.special as scsp
import random


class PySALEDistributionBase:
    """Base class for distributions.

    In PySALESetup it is not uncommon to want to feed specific sizes,
    or rotations, or radii, to a setup routine. These distribution
    classes provide this capability. Everything random always comes
    from a distribution and if it's not specified, it's probably a
    uniform one.
    """
    def __init__(self):
        self.mean = None
        self.median = None
        self.mode = None
        self.variance = None
        self.skew = None
        self.name = None

    def details(self):
        details = "distribution has the following properties:\n"
        details += f"Name: {self.name}"
        details += f"mean = {self.mean:2.3f}\n"
        details += f"median = {self.median:2.3f}\n"
        details += f"mode = {self.mode:2.3f}\n"
        details += f"variance = {self.variance:2.3f}\n"
        details += f"skewness = {self.skew:2.3f}\n"
        return details

    def frequency(self, x: float, bounds: Tuple[float, float]) -> float:
        """Get the frequency from the PDF over a specified interval.

        Integrates over the probability density function of the
        chosen distribution over a specified interval
        to return an estimated frequency. Limits MUST be provided
        in the form of `bounds`, which allows for uneven limits and is
        always applied as + and - the given value of x.

        Returns the probability DENSITY! this must be converted to a
        useful value outside of the function.
        Parameters
        ----------
        x : float
        bounds : Tuple[float, float]

        Returns
        -------
        frequency_density : float
        """
        f = np.float64(abs(self.cdf(x + bounds[1]) - self.cdf(x - bounds[0])))
        return f

    def cdf(self, x: float):
        """Base method for calculating the CDF at a specified point.

        Parameters
        ----------
        x : float

        Returns
        -------

        """
        raise NotImplementedError

    def random_number(self):
        """Base method for getting a random number by the distribution.

        Returns
        -------

        """
        raise NotImplementedError


class PySALEUniformDistribution(PySALEDistributionBase):
    """The uniform distribution.

    All values are equally likely.
    """
    def __init__(self, limits: Tuple[float, float]):
        """Creates a uniform distribution given a pair of limits.

        Parameters
        ----------
        limits : Tuple[float, float]
        """
        super().__init__()
        assert limits is not None, \
            "ERROR: function must have size limits (ind var)"
        self.name = 'Uniform'
        self.limits = limits
        self.mean = .5 * (limits[0] + limits[1])
        self.median = self.mean
        self.variance = (1. / 12.) * (limits[1] - limits[0]) ** 2.
        self.skew = 0.

    def cdf(self, x: float) -> float:
        """CDF for a uniform probability density function.

        Parameters
        ----------
        x : float

        Returns
        -------
        probability : float
        """
        minx = self.limits[0]
        maxx = self.limits[1]
        f = (x - minx) / (maxx - minx)
        if x < minx:
            f = 0.
        elif x >= maxx:
            f = 1.
        return f

    def random_number(self):
        """Generates a random number from the uniform distribution.

        Returns
        -------
        number : float
        """
        return np.random.uniform(*self.limits)


class PySALENormalDistribution(PySALEDistributionBase):
    """Normal distribution.

    """
    def __init__(self, mu: float, sigma: float):
        """Constructs a normal distribution given a mean and std.

        Parameters
        ----------
        mu : float
            mean
        sigma : float
            standard deviation
        """
        super().__init__()
        self.mu = mu
        self.sigma = sigma
        self.mean = mu
        self.median = mu
        self.mode = mu
        self.variance = sigma ** 2.
        self.skew = 0.
        self.name = 'Normal'

    def cdf(self, x) -> float:
        """CDF for a normal probability density function.

        Parameters
        ----------
        x : float

        Returns
        -------
        probability : float
        """
        mu = self.mu
        sigma = self.sigma
        f = .5 * (1. + scsp.erf((x - mu) / (sigma * np.sqrt(2.))))
        return f

    def random_number(self) -> float:
        """Generate a random number from the Normal distribution.

        Returns
        -------
        random_number : float
        """
        return np.random.normal(self.mean, np.sqrt(self.variance))


class PySALELogNormalDistribution(PySALEDistributionBase):
    """Lognormal distribution."""
    def __init__(self, mu: float, sigma: float):
        """Construct a Lognormal distribution from a mean and std.

        Parameters
        ----------
        mu : float
            mean
        sigma : float
            standard deviation
        """
        super().__init__()
        self.mu = mu
        self.sigma = sigma
        self.mean = np.exp(mu + 0.5 * sigma ** 2.)
        self.median = np.exp(mu)
        self.mode = np.exp(mu - sigma ** 2.)
        self.variance = (np.exp(sigma ** 2.) - 1.) \
                        * np.exp(2. * mu + sigma ** 2.)
        self.skew = (np.exp(sigma ** 2.) + 2.) \
                    * np.sqrt(np.exp(sigma ** 2.) - 1.)
        self.name = 'Lognormal'

    def cdf(self, x: float) -> float:
        """CDF for the log-normal probability density function.

        Parameters
        ----------
        x : float

        Returns
        -------
        probability : float
        """
        mu = self.mu
        sigma = self.sigma
        f = .5 + .5 * scsp.erf((np.log(x) - mu) / (sigma * np.sqrt(2.)))
        return f

    def frequency(self, x: float, bounds: Tuple[float, float]) -> float:
        """Frequency of the Lognormal PDF over a given range.

        Parameters
        ----------
        x : float
        bounds : Tuple[float, float]

        Returns
        -------
        frequency : float
        """
        assert x >= 0., "ERROR: Lognormal distribution only works " \
                        "for input greater than 0"
        if type(bounds) != list:
            warnings.warn('dx must be a list, set to list this time')
            bounds = [bounds * .5, bounds * .5]
        f = np.float64(abs(self.cdf(x + bounds[1]) - self.cdf(x - bounds[0])))
        return f

    def random_number(self) -> float:
        """Generate a random number from the Lognormal distribution.

        Returns
        -------
        random_number : float
        """
        return np.random.lognormal(self.mean, np.sqrt(self.variance))


class PySALEWeibull2Distribution(PySALEDistributionBase):
    """Weibull 2-parameter distribution.

    This distribution is typically used for Particle Size Distributions
    generated by grinding, milling, and crushing operations.
    """
    def __init__(self, lambda_: float, k: float):
        """Construct a Weibull 2-parameter distribution.

        Parameters
        ----------
        lambda_ : float
            The 'scale' of the distribution
        k : float
            The 'shape' of the distribution.
        """
        super().__init__()
        if lambda_ < 0:
            warnings.warn("lambda must be >= 0, not {:2.2f}; "
                          "setting to zero this time".format(lambda_))
            lambda_ = 0.
        if k < 0.:
            warnings.warn("k must be >= 0, not {:2.2f}; "
                          "setting to zero this time".format(k))
            k = 0.
        self.lambda_ = lambda_
        self.k = k
        self.mean = lambda_ * scsp.gamma(1. + 1. / k)
        self.median = lambda_ * (np.log(2.)) ** (1. / k)
        if k > 1:
            self.mode = lambda_ * ((k - 1) / k) ** (1. / k)
        else:
            self.mode = 0
        self.variance = (lambda_ ** 2.) * (scsp.gamma(1. + 2. / k)
                                           - (scsp.gamma(1. + 1. / k))
                                           ** 2.)
        self.skew = (scsp.gamma(1. + 3. / k) * lambda_ ** 3.
                     - 3. * self.mean * self.variance - self.mean ** 3.)
        self.skew /= self.variance ** (3. / 2.)
        self.name = 'Weibull 2-parameter'

    def cdf(self, x: float) -> float:
        """CDF for a Weibull 2-parameter distribution

        Parameters
        ----------
        x : float

        Returns
        -------
        probability : float
        """
        if x >= 0:
            f = 1. - np.exp(-((x / self.lambda_) ** self.k))
        else:
            f = 0.
        return f

    def random_number(self) -> float:
        """Generate a random number from the Weibull2 distribution.

        Returns
        -------
        random_number : float
        """
        r = np.random.uniform()
        x = self.lambda_ * (-np.log(r+1)) ** (1./self.k)
        return x


class PySALECustomDistribution(PySALEDistributionBase):
    """A user-defined distribution."""
    def __init__(self, cdf_function: Callable,
                 random_number_function: Callable):
        """Construct a user-defined distribution from two functions.

        Parameters
        ----------
        cdf_function : Callable
        random_number_function : Callable
        """
        super().__init__()
        assert callable(cdf_function), \
            "cdf function provided is not a callable!"
        assert callable(random_number_function), \
            "random number function provided is not a callable!"
        self.cdf = cdf_function
        self.random_number = random_number_function
        self.name = 'User-defined'

    def cdf(self, x: float):
        """Overwritten by the constructor"""
        raise NotImplementedError

    def random_number(self):
        """Overwritten by the constructor"""
        raise NotImplementedError

    def details(self):
        return "Custom distribution in use."


class PySALEDomain:
    """Domain class providing insertion methods and statistics.

    Create an instance from an existing PySALEObject and then use the
    class to insert objects in more complex ways than available on the
    standard object class. In particular `optimise_materials` is
    available as a way to distribute materials in a host object such
    that as few objects of the same material are in contact as possible.

    Examples
    --------


    >>> from PySALESetup import PySALEObject, PySALEDomain
    >>> main = PySALEObject([(0, 0), (0, 30), (30, 30), (30, 0)])
    >>> main.set_material(0)
    >>> domain = PySALEDomain(main)
    >>> circle = PySALEObject.generate_ellipse([0., 0.], 1, 1, 0.)
    >>> domain.fill_to_threshold_area(circle, 40)
    >>> domain.optimise_materials()
    >>> fig, ax = main.plot()
    >>> ax.set_aspect(1)
    >>> fig.savefig('PySALEDomain_example.png')

    """
    def __init__(self, domain_object: PySALEObject):
        """

        Parameters
        ----------
        domain_object
        """
        self.object: PySALEObject = domain_object

    def fill_with_random_grains_to_threshold(self,
                                             grain_object: PySALEObject,
                                             threshold_fill_percent: float,
                                             rotation_distribution:
                                             PySALEDistributionBase
                                             = None,
                                             size_distribution:
                                             PySALEDistributionBase
                                             = None,
                                             max_retries: int = 10
                                             ) -> float:
        """Fill host object to threshold fill percent.
        
        Parameters
        ----------
        grain_object : PySALEObject
            The filler object
        threshold_fill_percent : float
        rotation_distribution : PySALEDistributionBase
            The distribution for the rotation angles.
        size_distribution : PySALEDistributionBase
            The distribution for the grain sizes.
        max_retries : int
            If a grain fails to be placed how many 
            retries with new grains before giving up? Does not reset
            between grains.

        Returns
        -------
        float
            Inserted area
        """
        inserted_area, insertion_possible, threshold \
            = self._check_threshold_input(threshold_fill_percent)
        retries = 0
        while inserted_area <= threshold and retries <= max_retries:
            if rotation_distribution is not None:
                grain_object = self.randomly_rotate_object(
                    grain_object, rotation_distribution
                )
            if size_distribution is not None:
                grain_object = self.randomly_resize_object(
                    grain_object, size_distribution
                )
            insertion_possible = self.insert_randomly(grain_object)
            if insertion_possible is not True:
                retries += 1
                continue
            inserted_area = sum([c.area for c in self.object.children])
        return inserted_area

    @staticmethod
    def randomly_rotate_object(duplicated_grain_object: PySALEObject,
                               rotation_distribution:
                               PySALEDistributionBase = None) \
            -> PySALEObject:
        """Rotate the supplied object by a random amount.
        
        Parameters
        ----------
        duplicated_grain_object : PySALEObject
        rotation_distribution : PySALEDistributionBase

        Returns
        ------- 
        rotated_object : PySALEObject
        """
        if rotation_distribution is not None:
            angle = rotation_distribution.random_number()
        else:
            angle = 0.
        duplicated_grain_object = duplicated_grain_object.rotate(angle)
        return duplicated_grain_object

    @staticmethod
    def randomly_resize_object(grain_object: PySALEObject,
                               size_distribution:
                               PySALEDistributionBase = None,
                               area: bool = False) \
            -> PySALEObject:
        """Resize the supplied object by a random amount.
        
        The supplied distribution should return either areas or radii.
        Areas and radii are taken to be "equivalent" radii. I.e. the
        radius of the equivalent circle to the polygon with the same
        area.
        
        Parameters
        ----------
        grain_object : PySALEObject
        size_distribution : PySALEDistributionBase
        area : bool
            True if the distribution is returning an area, False if its
            an equivalent radius.

        Returns
        ------- 
        rotated_object : PySALEObject
        """
        if size_distribution is not None:
            rand_val = size_distribution.random_number()
        else:
            rand_val = grain_object.area if area \
                else grain_object.calculate_equivalent_radius()

        if area is True:
            factor = rand_val/grain_object.area
        else:
            factor = rand_val/grain_object.calculate_equivalent_radius()
        return grain_object.scale_object(factor, area)

    def _check_threshold_input(self, threshold_fill_percent: float):
        target_fraction = threshold_fill_percent / 100.
        threshold = target_fraction * self.object.area
        assert 0. < target_fraction < 1., f"Target area percentage " \
                                          f"must be between 0 and 100, " \
                                          f"exclusive;" \
            f" not {threshold_fill_percent}"
        inserted_area = sum([c.area for c in self.object.children])
        current_fraction = inserted_area / self.object.area
        assert current_fraction < target_fraction, \
            f"Target fraction ({target_fraction}) " \
            f"less than current fraction ({current_fraction})."
        insertion_possible = True
        return inserted_area, insertion_possible, threshold

    def insert_randomly(self, grain_object: PySALEObject,
                        max_attempts: int = 100) -> bool:
        """Insert object into the host at random locations until it fits

        "fits" means - "is not intersecting with any other objects."

        Parameters
        ----------
        grain_object : PySALEObject
        max_attempts : int
            The number of attempts the algorithm will make
            before giving up

        Returns
        -------
        success : bool
        """
        assert type(max_attempts) == int and max_attempts > 0,\
            f'Maximum number of attempts ' \
            f'MUST be a positive integer and not {max_attempts}'
        minx, miny, maxx, maxy = self.object.bounds
        if not self.object.has_children:
            intersects_with_other_shapes = False
        else:
            intersects_with_other_shapes = True
        is_within_domain = False
        mp = MultiPolygon(self.object.children)
        counter = 0
        success = False
        while intersects_with_other_shapes \
                or is_within_domain is not True:
            grain_object = self._move_object_to_random_coordinate_in_domain(grain_object, maxx, maxy, minx, miny)
            is_within_domain = self.object.contains(grain_object)
            intersects_with_other_shapes = mp.intersects(grain_object)
            counter += 1
            if counter >= max_attempts:
                warnings.warn(f'Max insertion attempts '
                              f'reached ({max_attempts}). '
                              f'Object can not be placed.')
                success = False
                break
            else:
                success = True
        if is_within_domain and not intersects_with_other_shapes:
            self.object.add_child(grain_object)
        return success

    @staticmethod
    def _move_object_to_random_coordinate_in_domain(grain_object,
                                                    maxx, maxy,
                                                    minx, miny) \
            -> PySALEObject:
        centroid = np.random.uniform(low=minx, high=maxx), \
                   np.random.uniform(low=miny, high=maxy)
        grain_object = grain_object.translate(*centroid)
        return grain_object

    def optimise_materials(self,
                           allowed_materials: List[int]
                           = (1, 2, 3, 4, 5, 6, 7, 8, 9)) -> None:
        """Redistribute material numbers to not be in contact.

        This function has the greatest success and is based on that
        used in JP Borg's work with CTH.

        Function to assign material numbers to each particle
        This function tries to optimise the assignments such
        that as few particles of the same material are in contact as
        possible. It works by creating a list of the Mth
        closest polygons, where "M" is the number of
        different materials being used.

        Then they are sorted into the order closest -> furthest.

        If the list contains all the same elements, as the
        allowed materials then there are necessarily no repeats
        and all material numbers are used up. So,
        => use the number of the particle furthest away.

        If there is at least one repeat, then at least one
        material number has not been used. Select the
        first remaining material number that is unused
        and assign that.

        Continue until all the particles are assigned.

        Parameters
        ----------
        allowed_materials : List[int]
            defaults to all materials except for void.

        Returns
        -------
        None
        """
        num_materials = len(set(allowed_materials))
        self._reset_domain_object_materials()

        for child in self.object.children:
            ordered_neighbours = sorted(self.object.children[1:],
                                        key=child.centroid.distance)
            assigned_materials = set([o._material
                                      for o, m in zip(ordered_neighbours,
                                                      set(allowed_materials))
                                      if o._material is not None])
            if len(assigned_materials) == num_materials:
                # All material numbers have been used up and no repeats!
                # (if there were repeats not all would have
                # been used...)
                new_material = random.choice(allowed_materials)
            else:
                missing_materials = [item
                                     for item in allowed_materials
                                     if item not in assigned_materials]
                new_material = random.choice(missing_materials)
            child.set_material(new_material)

    def _reset_domain_object_materials(self):
        for child in self.object.children:
            child._material = None





import os
import pathlib

PACKAGE_ROOT_DIRECTORY = pathlib.Path(os.path.dirname(__file__))
ASTEROID_TEMPLATE_PATH = PACKAGE_ROOT_DIRECTORY / 'input_templates' \
                         / 'asteroid_template.inp'
ADDITIONAL_TEMPLATE_PATH = PACKAGE_ROOT_DIRECTORY / 'input_templates' \
                         / 'additional_template.inp'
GRAIN_LIBRARY_PATH = PACKAGE_ROOT_DIRECTORY / 'grain_library_directory'


from PySALESetup.objects import PySALEObject, Velocity
from PySALESetup.functions import get_figure_from_ax
from collections import namedtuple
import numpy as np
from dataclasses import dataclass
from shapely.geometry import Point
from typing import Iterable, Optional, Tuple, Union, List, Dict
import matplotlib.pyplot as plt
from pathlib import Path
import warnings
import gzip
from enum import Enum
from mpl_toolkits.axes_grid1 import make_axes_locatable


class Region(Enum):
    NORTH = 1
    SOUTH = 2
    EAST = 3
    WEST = 4


ExtensionZoneFactor = namedtuple('ExtensionZoneFactors',
                                 ['multiplier', 'max_cell_size'])


class ExtensionZone:
    """iSALE Extension zone object.


    Extension zones can be bolted onto the main mesh (aka the
    high resolution zone). These can only have one material,
    a fixed velocity and a specific depth. Up to four can be
    added at a time and the order they appear in the PySALEMesh
    object is the order they will be applied.

    .. code-block::

                North
                  |
                  |
        West----MESH----East
                  |
                  |
                South
    """
    def __init__(self, depth: int, region: Region, cell_size: float,
                 factor: ExtensionZoneFactor = None):
        self.depth = depth
        self.region = region
        self.cell_size = cell_size
        if factor is None:
            factor = ExtensionZoneFactor(1, cell_size)
        self.factor = factor

    @property
    def length(self) -> float:
        """Physical length of the zone.

        Returns
        -------
        length : float
        """
        return self.calculate_zone_length()

    def calculate_zone_length(self) -> float:
        """Calculate physical length of the zone.

        Returns
        -------
        float
        """
        total_length = 0
        for i in range(self.depth):
            if self.cell_size < self.factor.max_cell_size:
                self.cell_size *= self.factor.multiplier
            total_length += self.cell_size
        return total_length


@dataclass
class Cell:
    """PySALEMesh cell dataclass object.

    Contains cell information including the physical centroid of a cell
    as a shapely.geometry.Point object; the indices of the cell in the
    mesh as integers; the material in the cell as an integer;
    the velocity of the cell as a pss Velocity object.
    """
    point: Point
    i: int
    j: int
    material: int = None
    velocity: Velocity = Velocity(0., 0.)


class PySALEMesh:
    """Mesh object of discrete cells which polygons can be projected on.

    Examples
    --------
    Once creating a polygon (or polygons) to represent your simulation
    it must then be projected onto/applied to this mesh object in order
    to save the resulting mesh to an iSALE input file.

    Here we create a void polygon that is 10 m x 15 m and populate
    it to 40% area fraction with circles of radius 1 m. Then we
    optimise the material distribution and apply the polygon to a
    mesh object. Then we can save the file.

    >>> from PySALESetup import PySALEObject, PySALEDomain, PySALEMesh
    >>> main = PySALEObject([(0, 0), (0, 15), (10, 15), (10, 0)])
    >>> main.set_material(0)
    >>> domain = PySALEDomain(main)
    >>> circle = PySALEObject.generate_ellipse([0., 0.], 1, 1, 0.)
    >>> domain.fill_with_random_grains_to_threshold(circle, 40)
    >>> domain.optimise_materials()
    >>> mesh = PySALEMesh(100, 150, cell_size=.1)
    >>> mesh.project_polygons_onto_mesh([main])
    >>> mesh.save()
    """
    def __init__(self,
                 x_cells: int,
                 y_cells: int,
                 cell_size: float = 2.e-6,
                 extension_zones: List[ExtensionZone] = None,
                 cylindrical_symmetry: bool = False,
                 collision_index: int = 0,
                 origin: Tuple[float, float] = (0., 0.)):
        """Discrete rectangular mesh construction.

        Parameters
        ----------
        x_cells : int
        y_cells : int
        cell_size : float
        extension_zones : Optional[List[ExtensionZone]]
        cylindrical_symmetry : bool
        collision_index : int
        origin : Tuple[float, float] - The origin for the
                 coordinate system.
        """
        self.x = x_cells
        self.y = y_cells
        self._origin = origin
        self.cell_size = cell_size
        self._x_range = None
        self._y_range = None
        self._cells = None
        self._material_meshes = None
        self._velocities = None
        self._extension_zones = extension_zones
        self._extension_factor = None
        self._y_physical_length = None
        self._x_physical_length = None
        self.cylindrical_symmetry = cylindrical_symmetry
        self._collision = collision_index

    def __str__(self):
        ezs = {Region.NORTH: 'N', Region.EAST: 'E',
               Region.WEST: 'W', Region.SOUTH: 'S'}
        extensions = [ezs[z.region] for z in self.extension_zones]
        s = '[' + ','.join(extensions) + ']'
        return f'PySALEMesh({self.x}, {self.y}, {self.cell_size}, ' \
               f'extension_zones={s}, origin={self.origin})'

    @property
    def origin(self) -> Tuple[int, int]:
        """The origin coordinate of the mesh.

        All dimensions are relative to this coordinate.
        Defaults to (0, 0). This coordinate is in the coordinate system
        of the mesh before the origin is applied. This has the 0, 0
        at the bottom-left corner of the *high-resolution* zone.

        Returns
        -------
        origin : Tuple[int, int]
        """
        return self._origin

    @classmethod
    def from_dimensions(cls, dimensions: Tuple[float, float],
                        cell_size: float,
                        extensions: Optional[List[ExtensionZone]] = None,
                        origin: Tuple[float, float] = (0., 0.)) \
            -> 'PySALEMesh':
        """Given high-res zone dimensions and cell size, return PySALEMesh.

        Parameters
        ----------
        dimensions : Tuple[float, float] X - Y Dimensions of the high-res
                     region in metres
        cell_size : float Dimension of a high-res cell in the mesh
        extensions : List[ExtensionZone] List of all the extension zones
                     that should be applied
        origin : Tuple[float, float] The coordinate to be considered the
                 origin. This coordinate is in the same coordinate
                 system as the default, where the origin is the bottom
                 left of the high-res zone.

        Returns
        -------
        PySALEMesh instance.
        """
        x_cells = round(dimensions[0] / cell_size)
        y_cells = round(dimensions[1] / cell_size)
        mesh = cls(x_cells, y_cells, cell_size,
                   extension_zones=extensions,
                   origin=origin)
        return mesh

    def _find_extension_factor(self):
        if self._extension_zones:
            assert all([e.cell_size == self.cell_size
                        for e in self._extension_zones]), \
                "All extension zones must have the same cell size!"
            self._extension_factor = self._extension_zones[0].factor
        else:
            self._extension_factor = \
                ExtensionZoneFactor(1., self.cell_size)

    @property
    def x_physical(self):
        """The physical x-length of the mesh.

        Returns
        -------
        length : float
        """
        if self._x_physical_length is None:
            self._populate_n_range()
        return self._x_physical_length

    @property
    def y_physical(self):
        """The physical y-length of the mesh.

        Returns
        -------
        length : float
        """
        if self._y_physical_length is None:
            self._populate_n_range()
        return self._y_physical_length

    @property
    def objresh(self) -> int:
        """iSALE input parameter; half the **height** of the mesh.

        Despite being the "horizontal" object resolution this refers to
        the height of the mesh. This is because there IS an OBJRESV
        but you only need to use it if your object does not cover
        the full width of the mesh. If no OBJRESV is present, its value
        defaults to OBJRESH. When using PySALESetup-created input files
        we never want anything less than the full width of the mesh so
        it is simpler to leave it out and use OBJRESH instead. In
        PySALESetup you can easily create objects of any size or shape
        you want!

        Notes
        -----
        If the number of cells is not divisible by 2, this property will
        guarantee that the returned value is rounded up, rather than
        down.

        E.g. a mesh width of 100 would return a value of 50.
        A mesh width of 99 would *also* return a value of 50.
        98 would return a value of 49, and so on.

        Returns
        -------
        objresh : int

        """
        if self.y % 2 == 0:
            objresh = int(self.y / 2)
        else:
            objresh = int((self.y // 2) + 1)
        return objresh

    @property
    def vertical_offset(self) -> int:
        """Half the vertical depth of the mesh, rounded down, in cells.

        Returns
        -------
        offset : int

        """
        if self.y % 2 == 0:
            offset = int(self.y / 2)
        else:
            offset = int((self.y-1) / 2)
        return offset

    @property
    def max_cell_size(self) -> float:
        """Return the maximum allowed cell size according to extensions.

        No extensions returns a max cell size identical to the
        mesh cell size.

        Returns
        -------
        max_cell_size : float
        """
        max_cell_size = self.cell_size
        if self.extension_zones:
            max_cell_size = self.extension_factor.max_cell_size
        return max_cell_size

    @property
    def collision_site(self) -> int:
        """The vertical collision location in the mesh, in cells.

        Defaults to 0.

        Returns
        -------
        collision_site : int

        """
        return self._collision

    @collision_site.setter
    def collision_site(self, value: int):
        self._collision = value

    @property
    def x_range(self) -> np.ndarray:
        """Array of the cell x-positions in the mesh.

        Returns
        -------
        x_range : float
        """
        if self._x_range is None:
            self._populate_n_range()
            self._set_origin()
        return self._x_range

    @property
    def y_range(self) -> np.ndarray:
        """Array of the cell y-positions in the mesh.

        Returns
        -------
        x_range : float
        """
        if self._y_range is None:
            self._populate_n_range()
            self._set_origin()
        return self._y_range

    def _set_origin(self):
        self._y_range -= self._origin[1]
        self._x_range -= self._origin[0]
        return

    def get_geometric_centre(self) -> Tuple[float, float]:
        """Return the geometric centre of the mesh in physical coords.

        Returns
        -------
        centre : Tuple[float, float]
        """
        x = np.ptp(self.x_range)*.5 + self.x_range[0]
        y = np.ptp(self.y_range)*.5 + self.y_range[0]
        return x, y

    @property
    def material_meshes(self) -> Dict[int, np.ndarray]:
        """Dictionary of numpy arrays representing material fill,
        indexed by material number.

        Returns
        -------
        meshes : Dict[int, np.ndarray]
        """
        if self._material_meshes is None:
            self._populate_material_meshes()
        return self._material_meshes

    @property
    def velocities(self) -> Dict[str, np.ndarray]:
        """Velocity arrays in the mesh in a dict indexed by axis.

        Returns
        -------
        velocities : Dict[str, np.ndarray]
        """
        if self._velocities is None:
            self._populate_velocities()
        return self._velocities

    def _populate_n_range(self):
        # funcs. they are a bit off. I think y-range works now
        # but x-range still broke. They should be combined probably
        y_length = self.y
        x_length = self.x
        self._y_physical_length = self.y * self.cell_size
        self._x_physical_length = self.x * self.cell_size
        zones = {zone.region: zone for zone in self.extension_zones}
        highres_xstart = 0
        highres_ystart = 0
        highres_xend = self.x
        highres_yend = self.y
        south_range = [-0.5 * self.cell_size]
        west_range = [-0.5 * self.cell_size]
        if Region.SOUTH in zones:
            zone = zones[Region.SOUTH]
            y_length += zone.depth
            highres_ystart = zone.depth
            highres_yend += zone.depth
            south_range = self._insert_extension_zone(zone)
        if Region.WEST in zones:
            zone = zones[Region.WEST]
            x_length += zone.depth
            highres_xstart = zone.depth
            highres_xend += zone.depth
            west_range = self._insert_extension_zone(zone)
        if Region.NORTH in zones:
            zone = zones[Region.NORTH]
            y_length += zone.depth
            highres_yend = highres_ystart + self.y + 1
            north_range = self._insert_extension_zone(zone)
        if Region.EAST in zones:
            zone = zones[Region.EAST]
            x_length += zone.depth
            highres_xend = highres_xstart + self.x + 1
            east_range = self._insert_extension_zone(zone)

        self._y_range = np.zeros((y_length))
        self._x_range = np.zeros((x_length))
        highres_yend_pos = (self.y+.5) * self.cell_size
        highres_xend_pos = (self.x+.5) * self.cell_size
        if Region.SOUTH in zones:
            self._y_range[:highres_ystart] = south_range
            highres_ystart_pos = np.amax(south_range)
            highres_yend_pos += highres_ystart_pos
        if Region.WEST in zones:
            self._x_range[:highres_xstart] = west_range
            highres_xstart_pos = np.amax(west_range)
            highres_xend_pos += highres_xstart_pos

        self._y_range[highres_ystart:highres_yend] = \
            self._generate_highres_zone(highres_yend,
                                        highres_ystart,
                                        south_range)
        self._x_range[highres_xstart:highres_xend] = \
            self._generate_highres_zone(highres_xend,
                                        highres_xstart,
                                        west_range)
        if Region.NORTH in zones:
            self._y_range[highres_yend-1:] = north_range
        if Region.EAST in zones:
            self._x_range[highres_xend-1:] = east_range
        return self._x_range, self._y_range

    def _generate_highres_zone(self,
                               highres_end,
                               highres_start,
                               range_):
        highres_zone = [np.amax(range_) + i * self.cell_size
                        for i in range(1, highres_end-highres_start+1)]
        return np.array(highres_zone)

    def _populate_velocities(self):
        x_cells = self.x_range.size
        y_cells = self.y_range.size
        self._velocities = {r: np.zeros((x_cells, y_cells))
                            for r in ['x', 'y']}

    def _populate_material_meshes(self):
        x_cells = self.x_range.size
        y_cells = self.y_range.size
        self._material_meshes = {i: np.zeros((x_cells, y_cells))
                                 for i in range(1, 9 + 1)}

    @property
    def cells(self) -> List[Cell]:
        """List of all Cell objects in the mesh.

        The mesh is represented by a collection of Cell objects,
        each of which represents a single cell in the mesh. These Cell
        objects are namedtuples containing all the information needed
        about that cell, including its indices, geometric centre,
        velocity, and material.

        Returns
        -------
        cells : List[Cell]
        """
        if self._cells is None:
            self._populate_cells()
        return self._cells

    def _populate_cells(self):
        self._cells = [Cell(Point(x, y), i, j, None, Velocity(0., 0.))
                       for i, x in enumerate(self.x_range)
                       for j, y in enumerate(self.y_range)]

    def project_polygons_onto_mesh(self,
                                   polygons: List[PySALEObject]) -> None:
        """Project a polygon (and all its children) onto the mesh.

        Method calls itself recursively on all children of the polygon.
        The children at the bottom of the hierachy get priority. Once
        a cell is populated with material, new material will NOT
        overwrite it.

        Parameters
        ----------
        polygons : List[PySALEObject]

        Returns
        -------
        None

        Examples
        --------
        Here we create a solid circle that is 5 m x 5 m and populate
        it to 40% area fraction with circles of radius 0.5 m. Then we
        optimise the material distribution and apply the polygon to a
        mesh object. Then we plot the result.

        >>> from PySALESetup import PySALEObject, PySALEDomain, PySALEMesh
        >>> import matplotlib.pyplot as plt
        >>> main = PySALEObject.generate_ellipse([5., 5.], 5., 5., 0.)
        >>> main.set_material(1)
        >>> domain = PySALEDomain(main)
        >>> circle = PySALEObject.generate_ellipse([0., 0.], .5, .5, 0.)
        >>> domain.fill_with_random_grains_to_threshold(circle, 40)
        >>> domain.optimise_materials([2, 3, 4, 5])
        >>> mesh = PySALEMesh(100, 100, cell_size=.1)
        >>> mesh.project_polygons_onto_mesh([main])
        >>> mesh.plot_materials()
        >>> plt.show()
        """
        for i, cell in enumerate(self.cells):
            if cell.material is None:
                self._project_polygons_onto_cell(cell, polygons)

    def _project_polygons_onto_cell(self, cell: Cell, polygons):
        for polygon in polygons:
            if cell.point.within(polygon):
                if polygon.children:
                    self._project_polygons_onto_cell(
                        cell,
                        polygon.children
                    )
                if cell.material is None:
                    self._fill_cell(cell, polygon)
                    break

    def _fill_cell(self, cell: Cell, geometry: PySALEObject):
        """Fill a mesh cell with the properties of a given polygon.

        Parameters
        ----------
        cell: Cell
        geometry : PySALEObject

        Returns
        -------
        None

        """
        if geometry.material == 0:
            self._void_cell(cell)
        else:
            self.material_meshes[geometry.material][cell.i, cell.j] = 1.
            self.velocities['x'][cell.i, cell.j] = geometry.velocity.x
            self.velocities['y'][cell.i, cell.j] = geometry.velocity.y
            cell.material = geometry.material
            cell.velocity = geometry.velocity

    def _void_cell(self, cell: Cell):
        """Fill a mesh cell with void.

        Parameters
        ----------
        cell : Cell

        Returns
        -------
        None
        """
        for number in self.material_meshes.keys():
            self.material_meshes[number][cell.i, cell.j] = 0.
        self.velocities['x'][cell.i, cell.j] = 0.
        self.velocities['y'][cell.i, cell.j] = 0.
        cell.material = 0
        cell.velocity = Velocity(0., 0.)

    def plot_cells(self, ax: plt.Axes = None):
        """Plot the cell centres of the mesh.

        Parameters
        ----------
        ax : plt.Axes

        Returns
        -------
        fig, ax : Tuple[plt.Axes, plt.figure]

        """
        ax, fig = get_figure_from_ax(ax)
        xi, yi = np.meshgrid(self.x_range, self.y_range)
        ax.scatter(xi, yi, marker='.', color='k')
        self._set_plot_lims_and_labels(ax)
        ax.set_title('Cell centres')
        return fig, ax

    def plot_materials(self, ax: plt.Axes = None,
                       cmap: str = 'rainbow') -> Tuple[plt.Figure,
                                                       plt.Axes]:
        """Plot the materials in the mesh using matplotlib.

        If no axes are provided, axes and a figure are made. Otherwise,
        the given axes are used and returned along with the associated
        figure object.

        Parameters
        ----------
        ax : plt.Axes
        cmap: str

        Returns
        -------
        fig, ax : Tuple[plt.Axes, plt.figure]

        Examples
        --------

        Here we construct a simple 2D meteorite impacting flat ground.
        Once our objects have been created and applied, we use
        plot_materials to view the mesh, although we need to use
        `plt.show()` to visualise the object you could just as easily
        save the figure instead.

        >>> from PySALESetup import PySALEObject
        >>> from PySALESetup import PySALEMesh
        >>> import matplotlib.pyplot as plt
        >>> impactor = PySALEObject.generate_ellipse([5., 8.], 2., 2., 0.)
        >>> impactor.set_material(1)
        >>> impactor.set_velocity(0. -1000.)
        >>> target = PySALEObject([(0, 0), (0, 6), (10, 6), (10, 0)])
        >>> target.set_material(3)
        >>> mesh = PySALEMesh(100, 100, cell_size=.1)
        >>> mesh.project_polygons_onto_mesh([impactor, target])
        >>> mesh.plot_materials()
        >>> plt.show()
        """
        ax, fig = get_figure_from_ax(ax)
        xi, yi = np.meshgrid(self.x_range, self.y_range)
        for i in range(1, 9+1):
            matter = np.copy(self.material_meshes[i])*i
            matter = np.ma.masked_where(matter == 0., matter)
            p = ax.pcolormesh(xi,
                              yi,
                              matter.T,
                              cmap=cmap,
                              vmin=1,
                              vmax=9,
                              shading='auto')
        self._set_plot_lims_and_labels(ax)
        self._add_colorbar(ax, p, 'Material No.')
        ax.set_title('Materials')
        return fig, ax

    @staticmethod
    def _add_colorbar(ax, graph_object, label):
        # create an axes on the right side of ax. The width of cax will be 5%
        # of ax and the padding between cax and ax will be fixed at 0.05 inch.
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        ax, fig = get_figure_from_ax(ax)
        cb = fig.colorbar(graph_object, cax=cax)
        cb.set_label(label)
        return cb

    def _set_plot_lims_and_labels(self, ax):
        ax.set_xlim(np.amin(self.x_range), np.amax(self.x_range))
        ax.set_ylim(np.amin(self.y_range), np.amax(self.y_range))
        ax.set_xlabel('$x$ [m]')
        ax.set_ylabel('$y$ [m]')

    def plot_velocities(self,
                        ax1: Optional[plt.Axes] = None,
                        ax2: Optional[plt.Axes] = None,
                        cmap: str = 'viridis') -> Tuple[plt.Figure,
                                                        plt.Figure,
                                                        plt.Axes,
                                                        plt.Axes]:
        """Plot the velocities of cells.

        If axes are provided they are used. If any are not provided,
        they are created. Either way the axes and figures are returned.

        Parameters
        ----------
        ax1 : Optional[plt.Axes]
        ax2 : Optional[plt.Axes]
        cmap : str

        Returns
        -------
        fig1, fig2, ax1, ax2 : Tuple[plt.Figure, plt.Figure,
                                     plt.Axes, plt.Axes]
        Examples
        --------

        >>> from PySALESetup import PySALEObject
        >>> from PySALESetup.mesh import PySALEMesh
        >>> import matplotlib.pyplot as plt
        >>> impactor = PySALEObject.generate_ellipse([5., 8.], 2., 2., 0.)
        >>> impactor.set_material(1)
        >>> impactor.set_velocity(0. -1000.)
        >>> target = PySALEObject([(0, 0), (0, 6), (10, 6), (10, 0)])
        >>> target.set_material(3)
        >>> mesh = PySALEMesh(100, 100, cell_size=.1)
        >>> mesh.project_polygons_onto_mesh([impactor, target])
        >>> mesh.plot_materials()
        >>> plt.show()

        """
        ax1, fig1 = get_figure_from_ax(ax1)
        ax2, fig2 = get_figure_from_ax(ax2)
        xi, yi = np.meshgrid(self.x_range, self.y_range)
        matter_sum = self.material_meshes[1]
        for i in range(2, 9+1):
            matter_sum += self.material_meshes[i]
        vx = np.ma.masked_where(matter_sum == 0., self.velocities['x'])
        vy = np.ma.masked_where(matter_sum == 0., self.velocities['y'])

        for ax, v in [(ax1, vx), (ax2, vy)]:
            p = ax.pcolormesh(xi, yi, v.T, cmap=cmap,
                              vmin=np.amin(v),
                              vmax=np.amax(v),
                              shading='auto')
            self._set_plot_lims_and_labels(ax)
            self._add_colorbar(ax, p, 'Velocity [m/s]')
        ax1.set_title('Velocity - x')
        ax2.set_title('Velocity - y')
        return fig1, fig2, ax1, ax2

    @staticmethod
    def _cell_to_row(cell: Cell, materials: Iterable[int]) -> str:
        material_columns = ['1.000' if cell.material == m
                            else '0.000' for m in materials]
        row = f'{cell.i} {cell.j} ' \
              + ' '.join(material_columns) \
              + f' {cell.velocity.x:.2f} {cell.velocity.y:.2f}\n'
        return row

    @property
    def material_numbers(self) -> List[int]:
        """List of non-zero materials in the mesh.

        Returns
        -------
        numbers : List[int]
        """
        return [key for key, value in self.material_meshes.items()
                if np.sum(value) > 0.]

    def save(self, file_name: Path = Path('./meso_m.iSALE'),
             compress: bool = False) -> None:
        """Save the current mesh to a meso_m.iSALE file.

        This compiles the integer indices of each cell,
        as well as the material in them. It saves all this to the file
        specified by the user, which defaults to ``meso_m.iSALE`` in
        the user's current directory.

        Parameters
        ----------
        file_name : Path
        compress : bool
            Compress the resulting text file using gunzip.
            Users are expected to add their own .gz file extension.
            If one is not present a UserWarning is raised.

        Returns
        -------
        None
        """
        cell_number = self.x * self.y
        if compress:
            if file_name.suffix != '.gz':
                raise NameError(f'Mesh is being compressed but file '
                                f'name "{file_name}" does not have a '
                                f'.gz extension.')
            with gzip.open(file_name, 'wt') as f:
                self._write_mesh_to_file(cell_number,
                                         f,
                                         self.material_numbers)
        else:
            with open(file_name, 'w') as f:
                self._write_mesh_to_file(cell_number,
                                         f,
                                         self.material_numbers)

    def _write_mesh_to_file(self,
                            cell_number: int,
                            file_object,
                            material_numbers: Union[List[int],
                                                    Tuple[int]]) \
            -> None:
        first_row = f'{cell_number}, {len(material_numbers)}\n'
        file_object.write(first_row)
        file_object.writelines([self._cell_to_row(cell,
                                                  material_numbers)
                                for cell in self.cells])

    def _insert_extension_zone(self, zone: ExtensionZone):
        factor = self.extension_factor.multiplier
        max_size = self.extension_factor.max_cell_size
        varying_cell_size = self.cell_size
        if zone.region in [Region.NORTH, Region.EAST]:
            half = 'north/east'
            if zone.region == Region.NORTH:
                position = self._y_physical_length + self.cell_size*.5
            else:
                position = self._x_physical_length + self.cell_size*.5
        else:
            half = 'south/west'
            position = -0.5*self.cell_size

        varying_cell_size, coord = self._create_extension_zone_coordinates(
            factor, max_size, [position], varying_cell_size, zone,
            half)

        if half == 'north/east':
            range_ = np.array(coord)
            if zone.region == Region.NORTH:
                self._y_physical_length += np.ptp(range_)
            else:
                self._x_physical_length += np.ptp(range_)
            return range_
        else:
            range_ = np.array(coord)
            return range_[::-1]

    def _insert_north_zone(self, zone: ExtensionZone):
        factor = self.extension_factor.multiplier
        max_size = self.extension_factor.max_cell_size
        varying_cell_size = self.cell_size
        position = self._y_physical_length + self.cell_size*.5

        varying_cell_size, y_coord = self._create_extension_zone_coordinates(
            factor, max_size, [position], varying_cell_size, zone,
            'north/east')

        north_y_range = np.array(y_coord)
        self._y_physical_length += np.ptp(north_y_range)
        return north_y_range

    def _insert_east_zone(self, zone: ExtensionZone):
        factor = self.extension_factor.multiplier
        max_size = self.extension_factor.max_cell_size
        varying_cell_size = self.cell_size
        position = self._x_physical_length + self.cell_size*.5

        varying_cell_size, x_coord = self._create_extension_zone_coordinates(
            factor, max_size, [position], varying_cell_size, zone,
            'north/east')

        east_x_range = np.array(x_coord)
        self._x_physical_length += np.ptp(east_x_range)
        return east_x_range

    def _insert_west_zone(self, zone: ExtensionZone):
        factor = self.extension_factor.multiplier
        max_size = self.extension_factor.max_cell_size
        varying_cell_size = self.cell_size
        position = -0.5*self.cell_size

        varying_cell_size, x_coord = self._create_extension_zone_coordinates(
            factor, max_size, [position], varying_cell_size, zone,
            'south/west')

        west_x_range = np.array(x_coord)
        return west_x_range[::-1]

    def _insert_south_zone(self, zone: ExtensionZone):
        factor = self.extension_factor.multiplier
        max_size = self.extension_factor.max_cell_size
        varying_cell_size = self.cell_size
        position = -0.5*self.cell_size

        varying_cell_size, y_coord = self._create_extension_zone_coordinates(
            factor, max_size, [position], varying_cell_size, zone,
            'south/west')
        south_y_range = np.array(y_coord)
        # south_y_range += abs(np.amin(south_y_range)) + varying_cell_size
        # self._y_physical_length += np.amax(south_y_range)
        return south_y_range[::-1]

    def _create_extension_zone_coordinates(self, factor: float,
                                           max_size: float,
                                           coord: List[float],
                                           varying_cell_size: float,
                                           zone: ExtensionZone,
                                           half: str):
        counter = 1
        position = coord[0]
        while counter < zone.depth:
            if varying_cell_size < max_size:
                varying_cell_size = counter * factor * self.cell_size
            else:
                varying_cell_size = max_size
            if half.lower() == 'south/west':
                position -= varying_cell_size
            elif half.lower() == 'north/east':
                position += varying_cell_size
            coord.append(position)
            counter += 1
        return varying_cell_size, coord

    @property
    def extension_zones(self) -> List[ExtensionZone]:
        """The extension zones applied to the mesh.

        Returns
        -------
        zones : List[ExtensionZone]
        """
        if self._extension_zones is not None:
            self._find_extension_factor()
            return self._extension_zones
        else:
            return []

    @property
    def extension_factor(self) -> ExtensionZoneFactor:
        """The ExtensionZoneFactor associated with this mesh.

        There can only be one extension factor associated with a mesh.
        When this property is called it also checks that the given
        extension zones don't have clashing properties.

        Returns
        -------
        factor : ExtensionZoneFactor
        """
        if self._extension_factor is None:
            self._find_extension_factor()
        return self._extension_factor





from setuptools import setup

version = {}
with open("./PySALESetup/_version.py") as fp:
    exec(fp.read(), version)

with open("./README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='PySALESetup',
    python_requires='>3.6, <3.10',
    version=version['__version__'],
    packages=['PySALESetup'],
    url='https://github.com/jgd10/PySALESetup',
    license='MIT',
    author='James Derrick',
    author_email='jgd10.github@gmail.com',
    description='iSALE setup package',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    package_data={
        'PySALESetup.grain_library_directory':
            ['grain_library_directory/*'],
        'PySALESetup.input_templates':
            ['input_templates/*']
    },
    install_requires=[
        'matplotlib~=3.4.3',
        'numpy~=1.21.2',
        'Shapely~=1.7.0',
        'scipy~=1.2.1'
    ]
)



import numbers
from typing import Optional, Tuple, Dict, Union
import matplotlib.pyplot as plt


def get_figure_from_ax(ax: Optional[plt.Axes] = None) -> Tuple[plt.Axes,
                                                               plt.Figure]:
    """Get the matplotlib figure from an Axes or create some if None.

    Parameters
    ----------
    ax : plt.Axes

    Returns
    -------
    (ax, fig) : Tuple[plt.Axes, plt.figure]
    """
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, aspect='equal')
    else:
        fig = ax.get_figure()
    return ax, fig


def _convert_input_to_fortran_strings(
        input_values: Dict[str, Union[str, float, int]]
) -> Dict[str, str]:
    new = {}
    for key, value in input_values.items():
        if isinstance(value, numbers.Number) and \
                not isinstance(value, int):
            value = f'{value:.6e}'.replace('e', 'D')
        else:
            value = f'{value}'
        new[key] = value
    return new
